#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.


import typer
from mlinfra.amplitude import amplitude_client
from mlinfra.terraform.terraform import Terraform
from mlinfra.utils.constants import TF_PATH
from mlinfra.utils.utils import run_command
from typing_extensions import Annotated

app = typer.Typer()


@app.command("plan", help="Plan the terraform stack config")
def plan(
    stack_config_path: Annotated[
        str, typer.Option("--config-file", help="Path to the stack configuration file")
    ],
    # args: Annotated[List[str], typer.Argument(help="Additional arguments to pass")],
):
    """
    Plan the terraform stack config.

    Args:
        stack_config_path (str): Path to the stack configuration file.
        args (List[str]): Additional arguments to pass to the terraform plan command.
    """
    targets_list = Terraform(stack_config_path=stack_config_path).plan()
    targets_list = ""

    run_command(["terraform", f"-chdir={TF_PATH}", "init"], capture_output=False)
    args = ["-lock=false", "-input=false", "-compact-warnings"]
    # args.append(targets_list)
    amplitude_client.send_event(
        amplitude_client.PLAN_EVENT,
        event_properties={"modules": targets_list},
    )
    run_command(
        ["terraform", f"-chdir={TF_PATH}", "plan"] + args + ["-out", "tfplan.binary"],
        capture_output=False,
    )


@app.command("apply", help="Apply the terraform stack config")
def apply(
    stack_config_path: Annotated[
        str, typer.Option("--config-file", help="Path to the stack configuration file")
    ],
    # args: Annotated[List[str], typer.Argument(help="Additional arguments to pass")],  # = ["-lock=false", "-input=false", "-compact-warnings"],
):
    """
    Apply the terraform stack config.

    Args:
        stack_config_path (str): Path to the stack configuration file.
        args (List[str]): Additional arguments to pass to the terraform apply command.
    """
    targets_list = Terraform(stack_config_path=stack_config_path).plan()
    targets_list = ""

    run_command(["terraform", f"-chdir={TF_PATH}", "init"], capture_output=False)
    args = ["-lock=false", "-input=false", "-compact-warnings"]
    # args.append(targets_list)
    amplitude_client.send_event(
        amplitude_client.APPLY_EVENT,
        event_properties={"modules": targets_list},
    )
    run_command(
        ["terraform", f"-chdir={TF_PATH}", "plan"] + args + ["-out", "tfplan.binary"],
        capture_output=False,
    )
    run_command(
        ["terraform", f"-chdir={TF_PATH}", "apply", "-auto-approve", "tfplan.binary"],
        capture_output=False,
    )


@app.command("destroy", help="Tear down the deployed terraform stack")
def destroy(
    stack_config_path: Annotated[
        str, typer.Option("--config-file", help="Path to the stack configuration file")
    ],
    # args: Annotated[List[str], typer.Argument(help="Additional arguments to pass")],
):
    """
    Run terraform for the config file with the given action and args.
    """
    targets_list = Terraform(stack_config_path=stack_config_path).plan()
    targets_list = ""

    run_command(["terraform", f"-chdir={TF_PATH}", "init"], capture_output=False)
    args = ["-auto-approve"]
    # args.append(targets_list)
    amplitude_client.send_event(
        amplitude_client.DESTROY_EVENT,
        event_properties={"modules": targets_list},
    )
    run_command(["terraform", f"-chdir={TF_PATH}", "destroy"] + args, capture_output=False)

    # elif action == "force-unlock":
    #     file_processor.force_unlock()
    #     action = f"plan {args} -lock=false"



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from enum import Enum


class DeploymentType(Enum):
    """
    Enumeration that represents different types of deployments.
    """

    # managed cloud providers
    CLOUD_NATIVE = "cloud_native"
    CLOUD_VM = "cloud_vm"
    KUBERNETES = "kubernetes"

    # local providers
    KIND = "kind"
    K3S = "k3s"
    MINIKUBE = "minikube"
    MICROK8S = "microk8s"
    DOCKER_COMPOSE = "docker_compose"
    DOCKER = "docker"



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import pytest
from mlinfra.stack_processor.stack_generator import StackGenerator


class TestStackGenerator:
    # Raises an exception if 'name', 'provider', 'deployment' or 'stack' keys are missing in 'stack_config'.
    def test_raises_exception_if_missing_keys(self):
        stack_config = {
            "provider": {"name": "aws", "region": "us-west-2"},
            "deployment": {"type": "cloud_vm"},
            "stack": [],
        }
        with pytest.raises(Exception):
            StackGenerator(stack_config)

    # TODO: Uncomment this and implement the feature
    # Raises an exception if the length of 'stack_name' attribute is greater than 37 characters.
    # def test_raises_exception_if_stack_name_exceeds_37_characters(self):
    #     stack_config = {
    #         "name": "a" * 38,
    #         "provider": {"name": "aws", "region": "us-west-2"},
    #         "deployment": {"type": "cloud_vm"},
    #         "stack": [],
    #     }
    #     with pytest.raises(Exception):
    #         StackGenerator(stack_config)

    # # Raises an exception if 'provider' key in 'stack_config' is not a valid CloudProvider enum value.
    def test_raises_exception_if_provider_key_is_invalid(self):
        stack_config = {
            "name": "test_stack",
            "provider": {"name": "invalid_provider", "region": "us-west-2"},
            "deployment": {"type": "cloud_vm"},
            "stack": [],
        }
        with pytest.raises(NotImplementedError):
            StackGenerator(stack_config)

    # TODO: think about moving validations in a better place
    # Raises an exception if 'deployment' key in 'stack_config' is not a valid DeploymentType enum value.
    # def test_raises_exception_if_deployment_key_is_invalid(self):
    #     stack_config = {
    #         "name": "test_stack",
    #         "provider": {"name": "aws", "region": "us-west-2"},
    #         "deployment": {"type": "invalid_deployment"},
    #         "stack": [],
    #     }
    #     with pytest.raises(ValueError):
    #         StackGenerator(stack_config).generate()



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
from importlib import resources

from mlinfra import modules
from mlinfra.stack_processor.provider_processor.provider import (
    AbstractProvider,
)
from mlinfra.utils.constants import TF_PATH


class AWSProvider(AbstractProvider):
    """
    Represents a provider for the AWS infrastructure.
    Args:
        stack_name (str): The name of the stack.
        config (dict): The configuration object containing the provider settings.
    Attributes:
        stack_name (str): The name of the stack.
        config (dict): The configuration object containing the provider settings.
        account_id (str): The AWS account ID.
        region (str): The AWS region.
        access_key (str): The AWS access key (optional).
        secret_key (str): The AWS secret key (optional).
        role_arn (str): The AWS role ARN (optional).
    """

    def __init__(self, stack_name: str, config: dict):
        super().__init__(stack_name=stack_name, config=config)
        # required
        self.account_id = config.get("account_id")
        self.region = config.get("region")

        # optional
        self.access_key = config.get("access_key", None)
        self.secret_key = config.get("secret_key", None)
        self.role_arn = config.get("role_arn", None)

    # TODO: refactor statefile name
    def get_statefile_name(self) -> str:
        """
        Returns the name of the statefile for the current stack and region.
        Returns:
            str: The name of the statefile.
        """
        return f"tfstate-{self.stack_name}-{self.region}"

    def configure_provider(self):
        """
        Configures the provider by updating the provider configuration file.
        It sets the AWS region, allowed account IDs, and default tags.
        It also adds a random provider.
        """
        with open(resources.files(modules) / "cloud/aws/provider.tf.json", "r") as data_json:
            with open(f"./{TF_PATH}/provider.tf.json", "w", encoding="utf-8") as tf_json:
                data = json.load(data_json)
                data["provider"]["aws"]["region"] = self.region
                data["provider"]["aws"]["allowed_account_ids"] = [self.account_id]
                data["provider"]["aws"]["default_tags"]["tags"]["region"] = self.region
                # data["provider"]["aws"]["default_tags"]["tags"][
                #     "stack"
                # ] = self.stack_config_path

                # TODO: Add assume role block if applicable

                # add random provider
                with open(
                    resources.files(modules)
                    / "cloud/aws/terraform_providers/random/provider.tf.json",
                    "r",
                ) as random_provider:
                    random_provider_json = json.load(random_provider)
                data["provider"].update(random_provider_json["provider"])

                json.dump(data, tf_json, ensure_ascii=False, indent=2)

        # TODO: check when to add this...

        # with open(absolute_project_root() / "modules/cloud/aws/data.tf.json", "r") as data_json:
        #     with open(f"./{TF_PATH}/data.tf.json", "w", encoding="utf-8") as tf_json:
        #         json_data = json.load(data_json)

        #         json_data["data"]["terraform_remote_state"] = {
        #             "parent": {
        #                 "backend": "s3",
        #                 "config": {
        #                     "bucket": self.get_statefile_name(),
        #                     "key": "ultimate-mlops-stack",
        #                     "use_lockfile": True,
        #                     "region": self.region,
        #                 },
        #             }
        #         }
        #         json.dump(json_data, tf_json, ensure_ascii=False, indent=2)



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json

import yaml
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.enums.deployment_type import DeploymentType
from mlinfra.stack_processor.stack_processor.stack import (
    AbstractStack,
)
from mlinfra.utils.constants import TF_PATH


class KubernetesStack(AbstractStack):
    """
    The KubernetesStack class is responsible for generating the
    Terraform configuration files for the Kubernetes stack.
    """

    def __init__(
        self,
        state_file_name: str,
        region: str,
        account_id: str,
        provider: CloudProvider,
        deployment_type: DeploymentType,
        stacks: yaml,
    ):
        """
        Constructor for the KubernetesStack class.

        :param provider: The cloud provider
        :type provider: Provider
        :param stacks: The stacks configuration
        :type stacks: yaml
        """
        super().__init__(
            state_file_name=state_file_name,
            region=region,
            account_id=account_id,
            provider=provider,
            deployment_type=deployment_type,
            stacks=stacks,
        )
        self.output = {"output": []}

    def process_stack_config(self):
        pass

    def process_stack_modules(self):
        if self.stacks:
            for module in self.stacks:
                # extracting stack type
                stack_type = [key for key in module.keys()][0]

                if "name" in module[stack_type]:
                    name = module[stack_type]["name"]
                else:
                    raise KeyError(f"No application assigned to the stack: {stack_type}")

                json_module = {"module": {name: {}}}

                # TODO: pickup right module source based on the deployment type

                json_module["module"][name]["source"] = (
                    f"./modules/applications/{self.deployment_type.value}/{stack_type}/{name}/tf_module"
                )

                # Reading inputs and outputs from the config file
                # placed in the applications/application folder and
                # adding them to json config for the application module
                application_config = self._read_config_file(
                    stack_type=stack_type, application_name=name
                )

                if application_config is not None:
                    if "inputs" in application_config and application_config["inputs"]:
                        inputs = {}
                        for input in application_config["inputs"]:
                            if not input["user_facing"]:
                                input_name = input["name"]
                                # handle the input from the yaml config
                                # if the input value is not a string
                                input_value = (
                                    input["default"]
                                    if input["default"] != "None"
                                    else "${ %s }" % f"{input['value']}"
                                )
                                inputs.update({input_name: input_value})
                            json_module["module"][name].update(inputs)

                    if "outputs" in application_config and application_config["outputs"]:
                        for output in application_config["outputs"]:
                            if output["export"]:
                                output_val = "${ %s }" % f"module.{name}.{output['name']}"

                                self.output["output"].append(
                                    {output["name"]: {"value": output_val}}
                                )

                # Checks if there are params in the config file which can be
                # passed to the application module. Params are checked against
                # the application module yaml file
                if "params" in module[stack_type]:
                    # TODO: throw error for a param not existing in the yaml config
                    params = {}
                    for key, value in module[stack_type]["params"].items():
                        for input in application_config["inputs"]:
                            if key == input["name"]:
                                if input["user_facing"]:
                                    params.update({key: value})
                                else:
                                    raise KeyError(f"{key} is not a user facing parameter")
                    json_module["module"][name].update(params)

                with open(
                    f"./{TF_PATH}/stack_{stack_type}.tf.json", "w", encoding="utf-8"
                ) as tf_json:
                    json.dump(json_module, tf_json, ensure_ascii=False, indent=2)
        else:
            print(
                """
                  No stacks found in the configuration file.
                  Please check the documentation for more information.
                """
            )
            # TODO: think about throwing an exception here
            # raise Exception("No stacks found in the configuration file")

    def process_stack_outputs(self):
        self.output["output"].append({"state_storage": {"value": self.state_file_name}})
        self.output["output"].append(
            {
                "providers": {
                    "value": {
                        "aws": {
                            "region": self.region,
                            "account_id": self.account_id,
                        }
                    }
                }
            }
        )
        with open(f"./{TF_PATH}/output.tf.json", "w", encoding="utf-8") as tf_json:
            json.dump(self.output, tf_json, ensure_ascii=False, indent=2)

    def process_stack_inputs(self):
        pass

    def _common_service_input(self) -> any:
        return {
            "variable": [
                {"region": [{"type": "string", "default": self.region}]},
                {"account_id": [{"type": "string", "default": self.account_id}]},
            ]
        }

    def _user_input(self) -> any:
        user_input = []
        for v in self.stack_config["input_variables"]:
            input = {v["name"]: [{"type": "string"}]}
            if "default" in v:
                input[v["name"]][0]["default"] = v["default"]
            user_input.append(input)
        return user_input

    def _default_config_input(self) -> any:
        user_input = []
        if "environments" in self.stack_config:
            for env in self.stack_config["environments"]:
                if self._is_env_match(env["name"]):
                    if "variables" in env:
                        for k in env["variables"]:
                            user_input.append(
                                {
                                    k: [
                                        {
                                            "type": "string",
                                            "default": env["variables"][k],
                                        }
                                    ]
                                }
                            )
                    break
        return user_input

    def prepare_input(self):
        if "input_variables" in self.stack_config:
            json_output = {"variable": []}
            if self.is_stack_component:
                json_output = self._common_service_input()

            json_output["variable"].extend(self._user_input())
            json_output["variable"].extend(self._default_config_input())

            with open(f"./{TF_PATH}/variable.tf.json", "w", encoding="utf-8") as tf_json:
                json.dump(json_output, tf_json, ensure_ascii=False, indent=2)

    def generate(self):
        # logger.info("Processing Kubernetes Infrastructure")
        self.process_stack_config()
        self.process_stack_inputs()
        self.process_stack_modules()
        self.process_stack_outputs()


if __name__ == "__main__":
    pass



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
from abc import ABC, abstractmethod
from importlib import resources
from typing import Any, Dict, Union

import yaml
from mlinfra import modules
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.enums.deployment_type import DeploymentType


class AbstractStack(ABC):
    """
    Abstract class for all stacks.
    """

    def __init__(
        self,
        state_file_name: str,
        region: str,
        account_id: str,
        provider: CloudProvider,
        deployment_type: DeploymentType,
        stacks: yaml,
    ):
        """
        Initializes the stack.

        Args:
            state_file_name (str): The name of the state file for the stack.
            region (str): The region where the stack will be deployed.
            account_id (str): The ID of the account associated with the stack.
            provider (CloudProvider): The cloud provider for the stack.
            deployment_type (DeploymentType): The type of deployment for the stack.
            stacks (yaml): The stack configuration in YAML format.
        """
        self.state_file_name = state_file_name
        self.region = region
        self.account_id = account_id
        self.provider = provider
        self.deployment_type = deployment_type
        self.stacks = stacks

    def _read_config_file(
        self, stack_type: str, application_name: str, extension: str = "yaml"
    ) -> Union[Dict, Any]:
        """
        Reads the config file for the application and returns the config
        as a JSON object for application config and YAML for stack config.

        Args:
            stack_type (str): The type of the stack.
            application_name (str): The name of the application.
            extension (str, optional): The extension of the config file.

        Returns:
            Union[Dict, Any]: The configuration as a JSON object or YAML.
        """
        if self.provider.value == CloudProvider.LOCAL.value:
            file_path = (
                resources.files(modules)
                / f"applications/{self.provider.value}/{stack_type}/{application_name}/{application_name}.{extension}"
            )
        else:
            file_path = (
                resources.files(modules)
                / f"applications/{self.deployment_type.value}/{stack_type}/{application_name}/{application_name}_{self.deployment_type.value}.{extension}"
            )
        with open(file_path, "r", encoding="utf-8") as config_file:
            if extension == "yaml":
                return yaml.safe_load(config_file.read())
            else:
                return json.loads(config_file.read())

    @abstractmethod
    def process_stack_config(self):
        """
        Process the stack configuration and validates the config.
        """
        pass

    @abstractmethod
    def process_stack_inputs(self):
        """
        Process the stack inputs and validates the inputs.
        """
        pass

    @abstractmethod
    def process_stack_modules(self):
        """
        Generates the stack modules Terraform JSON configuration
        and validates the modules against the stack config and parameters.
        """
        pass

    @abstractmethod
    def process_stack_outputs(self):
        """
        Generates the required stack outputs.
        """
        pass

    @abstractmethod
    def generate(self):
        """
        Generates the stack configuration.
        """
        pass



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
from importlib import resources

from mlinfra import modules
from mlinfra.stack_processor.provider_processor.provider import (
    AbstractProvider,
)
from mlinfra.utils.constants import TF_PATH


class LocalProvider(AbstractProvider):
    """
    Represents provider for Local provider clusters
    """

    def __init__(self, stack_name: str, config: dict, deployment: dict):
        """
        Inits the Local Provider
        """
        super().__init__(stack_name, config)
        self.deployment = deployment

    def configure_provider(self):
        """
        Configures Local provider
        """
        with open(
            resources.files(modules) / f"local/{self.deployment['type']}/provider.tf.json", "r"
        ) as data_json:
            with open(f"./{TF_PATH}/provider.tf.json", "w", encoding="utf-8") as tf_json:
                data = json.load(data_json)
                json.dump(data, tf_json, ensure_ascii=False, indent=2)



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
import os
from importlib import resources
from typing import Tuple

# import hashlib
import boto3
import yaml
from botocore.config import Config
from mlinfra import modules
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.stack_processor.stack_generator import StackGenerator
from mlinfra.terraform.state_helper import StateHelper
from mlinfra.utils.constants import TF_LOCAL_STATE_PATH, TF_PATH
from mlinfra.utils.utils import (
    check_docker_installed,
    check_kind_installed,
    check_terraform_installed,
    clean_tf_directory,
    create_symlinks,
)


class Terraform:
    """
    This class is responsible for running terraform commands
    This file is responsible for configuring and running terraform commands
    Plan function undergoes the following steps:
    - checking if terraform is installed
    - checking if following versions are correct:
        - terraform
        - mlops_cli
    - checking if the mlinfra config file exists
    - clean .mlops_infra folder
    - check cloud credentials
    - check if region has three AZs
    - Generate terraform state storage
    - loop through items in config file
        - generate list of modules (should come from stack processor module)
        - create a different apply function for different deployment types
            - order modules in order of application
            - apply terraform modules with -target
    """

    def __init__(self, stack_config_path: str):
        self.stack_config_path = stack_config_path

    def check_config_file_exists(self):
        """This function is responsible for checking if the config file exists"""
        if not os.path.isfile(self.stack_config_path):
            raise FileNotFoundError(f"The file {self.stack_config_path} does not exist.")

        with open(self.stack_config_path, "r") as stream:
            try:
                data = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
                raise ValueError(f"{exc}")

        # check if all required keys are present in the config file
        required_keys = ["name", "provider", "deployment", "stack"]

        if not all(key in data for key in required_keys):
            missing_keys = [key for key in required_keys if key not in data]
            raise ValueError(f"The following keys are missing: {', '.join(missing_keys)}")

    def clean_ml_infra_folder(self, delete_dir: bool = True):
        """
        This function is responsible for cleaning the .mlops_infra folder
        if delete_dir is set to true, the folder will be deleted
        """
        if delete_dir:
            clean_tf_directory()
        else:
            print("The param delete_dir is set as false, skipping the directory deletion")

    def read_stack_config(self) -> yaml:
        # clean the generated files directory
        clean_tf_directory()

        # create the stack folder
        os.makedirs(TF_PATH, mode=0o777)
        create_symlinks(resources.files(modules), TF_PATH + "/modules")

        # TODO: generate hash of the stack config to not generate config all the time
        # sha256_hash = hashlib.sha256()

        # read the stack config file
        try:
            with open(self.stack_config_path, "r") as stack_config:
                config = yaml.safe_load(stack_config.read())

            # with open(stack_config, "rb") as stack_config_file:
            #     # Read and update hash in chunks of 4K
            #     for byte_block in iter(lambda: stack_config_file.read(4096), b""):
            #         sha256_hash.update(byte_block)
            #         stack_file_digest = sha256_hash.hexdigest()
            #
            # Check if a folder with the value of stack_file_digest already exists
            # if not, create a folder inside the TF_PATH with stack_file_digest as the name
            # os.makedirs(f"{TF_PATH}/{stack_file_digest}", mode=0o777)

            return config

        except FileNotFoundError:
            raise FileNotFoundError(f"Stack config file not found: {self.stack_config_path}")

    # TODO: write getters for state file name and region
    def process_config_file(self) -> Tuple[str, str, str, str]:
        """This function is responsible for processing the config file"""
        stack_processor = StackGenerator(stack_config=self.read_stack_config())
        stack_processor.generate()

        return (
            stack_processor.get_stack_name(),
            stack_processor.get_state_file_name(),
            stack_processor.get_region(),
            stack_processor.get_provider(),
        )

    def check_cloud_credentials(self):
        """This function is responsible for checking if the cloud credentials are present"""
        try:
            if not boto3.Session().get_credentials():
                raise ValueError("AWS credentials not found.")
        except Exception as e:
            raise ValueError(f"An error occurred while checking AWS credentials: {str(e)}")

    def check_region_has_three_azs(self, aws_region: str = "eu-central-1"):
        """This function is responsible for checking if the region has three availability zones"""
        ec2 = boto3.client("ec2", config=Config(region_name=aws_region))
        response = ec2.describe_availability_zones(
            Filters=[{"Name": "zone-type", "Values": ["availability-zone"]}]
        )
        azs = [
            zone["ZoneName"]
            for zone in response["AvailabilityZones"]
            if zone["RegionName"] == aws_region
        ]
        if len(azs) < 3:
            raise ValueError(
                f"""
                The region {aws_region} has less than three availability zones.
                You configured {aws_region}, which only has the availability zones: {azs}.
                Please choose a different region.
                """
            )

    # TODO: move this out of this class
    def check_terraform_state_storage(self, state_name: str, aws_region: str):
        """This function is responsible for checking if the terraform state storage is present"""
        state_helper = StateHelper(
            state=state_name,
            region=aws_region,
        )

        state_helper.manage_aws_state_storage()

    def check_local_state_storage(self, state_name: str):
        """
        This function is responsible for checking if the local state storage is present
        """
        os.makedirs(f"{TF_LOCAL_STATE_PATH}/{state_name}", mode=0o777, exist_ok=True)

    def generate_modules_list(self):
        """
        This function is responsible for generating a list of modules
        that will be applied via the -target function of terraform
        """
        modules_list = []
        for file in os.listdir(TF_PATH):
            if file.endswith(".tf.json"):
                with open(os.path.join(TF_PATH, file)) as json_file:
                    data = json.load(json_file)
                    if "module" in data:
                        key, _ = data["module"].popitem()
                        modules_list.append(f"module.{key}")
        return modules_list

    def generate_terraform_config(self) -> Tuple[str, str, str, str]:
        """This function is responsible for generating the terraform config file"""
        check_terraform_installed()
        # TODO: perform this after the cli package has been released
        # self.check_mlinfra_cli_installed()
        self.check_config_file_exists()
        self.clean_ml_infra_folder()
        stack_name, state_name, aws_region, provider_type = self.process_config_file()
        return stack_name, state_name, aws_region, provider_type

    def plan(self) -> str:
        """
        This function is responsible for running terraform plan command
        along with a couple of other preliminary checks.
        """
        stack_name, state_name, aws_region, provider_type = self.generate_terraform_config()
        if provider_type is not CloudProvider.LOCAL:
            self.check_cloud_credentials()
            self.check_region_has_three_azs(aws_region=aws_region)
            self.check_terraform_state_storage(state_name=state_name, aws_region=aws_region)
        else:
            check_kind_installed()
            check_docker_installed()
            # state name is set to stack as in local deployment, there is no
            # region and hence no state name.
            self.check_local_state_storage(state_name=stack_name)
        modules_list = self.generate_modules_list()
        return "".join(f" -target={item}" for item in modules_list)



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import mlinfra.cli.terraform as terraform
import mlinfra.cli.utilities as utilities
import typer

cli = typer.Typer(
    name="mlinfra",
    help="Welcome to mlinfra",
    add_completion=False,
    pretty_exceptions_show_locals=False,
    no_args_is_help=True,
    short_help=["h"],
)
cli.add_typer(terraform.app, name="terraform", help="Terraform tasks on config file")
cli.add_typer(
    utilities.app,
    name="utils",
    help="Utilities to help users perform actions on config file",
    rich_help_panel="Utilities",
)


if __name__ == "__main__":
    cli()



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
import os
import re
import shutil

# TODO: Update this section to run it more secure and
# remove the comment
import subprocess  # nosec
import sys
from typing import List

from rich.traceback import install as rich_tr_install

from .constants import TF_PATH


def clean_tf_directory():
    if os.path.isdir(TF_PATH):
        shutil.rmtree(TF_PATH)


def generate_tf_json(module_name: str, json_module: json):
    with open(f"./{TF_PATH}/{module_name}.tf.json", "w", encoding="utf-8") as tf_json:
        json.dump(json_module, tf_json, ensure_ascii=False, indent=2)


def check_terraform_installed():
    try:
        # TODO: Update this section to run it more secure and
        # remove the comment
        result = run_command(["terraform", "--version"])

        # Extract the version information from the output
        output_lines = result.stdout.split("\n")
        for line in output_lines:
            if line.startswith("Terraform"):
                version_info = line.split()[1]
                print(f"Terraform is installed. Version: {version_info}")

    except subprocess.CalledProcessError as e:
        print(f"Error: {e}")
        print("Terraform is not installed or an error occurred.")
        print("Please install Terraform and try again.")
        exit(1)


# TODO: create symlink of only respective provider for the deployment type, not everything.
def create_symlinks(source: str, destination: str) -> None:
    """
    Create symlinks for all.tf and .tpl files in the source directory to the destination directory
    This also skips files starting with a dot (.)
    :param source: source directory
    :param destination: destination directory

    :return: None
    """
    if not os.path.exists(destination):
        os.makedirs(destination)

    for item in os.listdir(source):
        if item.startswith("."):
            continue
        source_item = os.path.join(source, item)
        destination_item = os.path.join(destination, item)

        if os.path.isdir(source_item):
            create_symlinks(source_item, destination_item)
        elif os.path.isfile(source_item) and (
            source_item.endswith(".tf")
            or source_item.endswith(".tpl")
            or re.search(r".*values\.yaml", source_item)
        ):
            if not os.path.exists(destination_item):
                os.symlink(source_item, destination_item)


def terraform_tested_version():
    # TODO: read from a file and then populate this field
    return "1.10.2"


# def _verify_aws_cloud_credentials(self) -> None:
#     ensure_installed("aws")
#     try:
#         aws_caller_identity = boto3.client("sts").get_caller_identity()
#         configured_aws_account_id = aws_caller_identity["Account"]
#         required_aws_account_id = self.root().providers["aws"]["account_id"]
#         if required_aws_account_id != configured_aws_account_id:
#             raise UserErrors(
#                 "\nSystem configured AWS Credentials are different from the ones being used in the "
#                 f"Configuration. \nSystem is configured with credentials for account "
#                 f"{configured_aws_account_id} but the config requires the credentials for "
#                 f"{required_aws_account_id}."
#             )
#     except NoCredentialsError:
#         raise UserErrors(
#             "Unable to locate credentials.\n"
#             "Visit `https://docs.aws.amazon.com/sdk-for-java/v1/developer-guide/setup-credentials.html` "
#             "for more information."
#         )
#     except ClientError as e:
#         raise UserErrors(
#             "The AWS Credentials are not configured properly.\n"
#             f" - Code: {e.response['Error']['Code']} Error Message: {e.response['Error']['Message']}"
#             "Visit `https://docs.aws.amazon.com/sdk-for-java/v1/developer-guide/setup-credentials.html` "
#             "for more information."
#         )


def run_command(
    command: List[str],
    universal_newlines: bool = True,
    capture_output: bool = True,
    text: bool = True,
    check: bool = True,
    timeout: float | None = None,
) -> List[str]:
    """
    This function runs a command in the subprocess and returns the result.

    Args:
        command: List of command arguments to run.
        universal_newlines: Whether to use universal newlines mode.
        capture_output: Whether to capture the output of the command.
        text: Whether to treat the output as text.
        check: Whether to raise an exception if the command exits with a non-zero status.
        timeout: Timeout in seconds for the command.

    Returns:
        The result of the command execution.
    """
    return subprocess.run(  # nosec
        command,
        universal_newlines=universal_newlines,
        capture_output=capture_output,
        text=text,
        check=check,
        timeout=timeout,
    )


def check_docker_installed():
    """
    This function is responsible for checking if docker is installed
    """
    try:
        version = run_command(["docker", "--version"])
        if not version:
            return 1
    except Exception as e:
        return f"An error occurred while checking the docker version. {str(e)}"


def check_kind_installed():
    """
    This function is responsible for checking if kind is installed
    """
    try:
        version = run_command(["kind", "--version"])
        if not version:
            return 1
    except Exception as e:
        return f"An error occurred while checking the kind version. {str(e)}"


def check_minikube_installed():
    """
    This function is responsible for checking if minikube cli is installed
    """
    try:
        version = run_command(["minikube", "version"])
        if not version:
            return 1
    except Exception as e:
        return f"An error occurred while checking the minikube version. {str(e)}"


def safe_run(func):  # type: ignore
    def func_wrapper(*args, **kwargs):  # type: ignore
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if hasattr(sys, "_called_from_test"):
                raise e
            else:
                print(e)
                return None

    return func_wrapper


def setup_rich_traceback(debug_mode: bool = False):
    if debug_mode:
        rich_tr_install(show_locals=True)
    else:
        rich_tr_install(show_locals=False, suppress=[subprocess])



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from abc import ABC, abstractmethod


class AbstractProvider(ABC):
    """
    Abstract class for providers
    """

    def __init__(self, stack_name: str, config: dict):
        """
        Initializes the AbstractProvider object with the provided stack name and configuration object.

        Args:
            stack_name (str): The name of the stack.
            config (dict): The configuration for the provider.
        """
        self.stack_name = stack_name
        self.config = config

    @abstractmethod
    def configure_provider(self):
        """
        Abstract method that needs to be implemented by subclasses to configure the provider.
        """
        pass



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import typer
from mlinfra.amplitude import amplitude_client
from mlinfra.terraform.terraform import Terraform
from mlinfra.utils.constants import TF_PATH
from mlinfra.utils.utils import run_command
from typing_extensions import Annotated

app = typer.Typer()


@app.command(
    "generate-terraform-config", help=f"Generates the terraform config in the {TF_PATH} folder path"
)
def generate_terraform_config(
    stack_config_path: Annotated[
        str, typer.Option("--config-file", help="Path to the stack configuration file")
    ],
):
    f"""
    Generates the terraform config in the {TF_PATH} folder path

    Args:
        stack_config_path (str): Path to the stack configuration file.
    """
    modules = Terraform(stack_config_path).plan()
    current_properties = {"modules": modules}

    amplitude_client.send_event(
        amplitude_client.START_GEN_TERRAFORM_EVENT,
        event_properties=current_properties,
    )

    run_command(["terraform", f"-chdir={TF_PATH}", "init"], capture_output=False)
    print(
        f"""
            Terraform config has been generated in the {TF_PATH} folder.
        """
    )
    amplitude_client.send_event(
        amplitude_client.FINISH_GEN_TERRAFORM_EVENT,
        event_properties=current_properties,
    )


@app.command("estimate-cost", help="Estimate the cost of your stack config")
def estimate_cost(
    stack_config_path: Annotated[
        str, typer.Option("--config-file", help="Path to the stack configuration file")
    ],
):
    """
    Estimate cost of the contents of the config file.

    Args:
        stack_config_path (str): Path to the stack configuration file.
    """
    Terraform(stack_config_path).plan()

    run_command(["terraform", f"-chdir={TF_PATH}", "init"], capture_output=False)

    run_command(
        [
            "terraform",
            f"-chdir={TF_PATH}",
            "plan",
            "-no-color",
            "-lock=false",
            "-input=false",
            "-compact-warnings",
            "-out",
            "tfplan.binary",
        ],
        capture_output=False,
    )

    run_command(
        [
            "sh",
            "-c",
            f"terraform -chdir={TF_PATH} show -no-color -json tfplan.binary > {TF_PATH}/plan.json",
        ],
        capture_output=False,
    )

    run_command(
        ["infracost", "diff", "--show-skipped", "--no-cache", "--path", f"{TF_PATH}/plan.json"],
        capture_output=False,
    )
    amplitude_client.send_event(
        amplitude_client.COST_ESTIMATOR_EVENT,
        event_properties={},
    )



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.enums.deployment_type import DeploymentType
from mlinfra.stack_processor.deployment_processor.cloud_vm_deployment import (
    CloudVMDeployment,
)
from mlinfra.stack_processor.deployment_processor.kind_deployment import KindDeployment
from mlinfra.stack_processor.deployment_processor.kubernetes_deployment import (
    KubernetesDeployment,
)
from mlinfra.stack_processor.deployment_processor.minikube_deployment import MiniKubeDeployment
from mlinfra.stack_processor.provider_processor.aws_provider import (
    AWSProvider,
)
from mlinfra.stack_processor.provider_processor.local_provider import LocalProvider
from mlinfra.stack_processor.stack_processor.cloud_vm_stack import (
    CloudVMStack,
)
from mlinfra.stack_processor.stack_processor.kubernetes_stack import (
    KubernetesStack,
)
from mlinfra.stack_processor.stack_processor.local_stack import LocalStack


class StackGenerator:
    """
    A class that generates and configures infrastructure stacks based on the provided stack configuration.

    Attributes:
        stack_config (dict): The stack configuration provided to the StackGenerator object.
        stack_name (str): The name of the stack.
        account_id (str): The account ID associated with the stack.
        provider (str): The cloud provider for the stack.
        deployment_type (str): The type of deployment (cloud infrastructure or Kubernetes).
        state_file_name (str): The name of the state file.
        is_stack_component (bool): A flag indicating if the stack is a component of a larger stack.
        output (dict): The output configuration for the stack.
    """

    def __init__(self, stack_config):
        """
        Initializes the StackGenerator object with the provided stack configuration.

        Args:
            stack_config (dict): The stack configuration provided to the StackGenerator object.

        Raises:
            Exception: If the stack configuration is missing any required components.
        """
        self.stack_config = stack_config
        self.stack_name = ""
        self.account_id = ""
        self.provider = "aws"
        self.deployment_type = ""
        self.state_file_name = ""
        self.is_stack_component = True
        self.output = {"output": []}

        if (
            not self.stack_config["name"]
            or "provider" not in self.stack_config
            or "deployment" not in self.stack_config
            or "stack" not in self.stack_config
        ):
            raise Exception("Stack config component is missing")

        # this has to be done now as the stack config is read
        # and the provider details are needed for the state file name
        # and the region
        # TODO: Throw an error if the stack name exceeds 37 characters
        self.stack_name = self.stack_config["name"]
        self.region = self.stack_config["provider"].get("region", "")
        self.provider = self.configure_provider()

    # TODO: refactor statefile name
    def get_state_file_name(self):
        """
        Generates the state file name based on the stack name and region.

        Returns:
            str: The state file name.
        """
        self.state_file_name = f"tfstate-{self.stack_name}-{self.region}"
        return self.state_file_name

    def get_region(self) -> str:
        """
        Returns the region specified in the stack configuration.

        Returns:
            str: The region.
        """
        return self.region

    def get_provider(self) -> str:
        """
        Returns the provider type

        Returns:
            str: The provider type.
        """
        return self.provider

    def get_stack_name(self) -> str:
        """
        Returns the stack name.

        Returns:
            str: The stack name.
        """
        return self.stack_name

    def generate(self):
        """
        Generates and configures the infrastructure stacks based on the deployment type.
        """
        deployment_type = self.stack_config["deployment"]["type"]

        if deployment_type == DeploymentType.CLOUD_VM.value:
            CloudVMDeployment(
                stack_name=self.stack_name,
                provider=CloudProvider(self.stack_config["provider"]["name"]),
                region=self.region,
                deployment_config=self.stack_config["deployment"],
            ).configure_deployment()

            CloudVMStack(
                state_file_name=self.state_file_name,
                region=self.region,
                account_id=self.account_id,
                provider=self.provider,
                deployment_type=DeploymentType.CLOUD_VM,
                stacks=self.stack_config["stack"],
            ).generate()

        elif deployment_type == DeploymentType.KUBERNETES.value:
            KubernetesDeployment(
                stack_name=self.stack_name,
                provider=CloudProvider(self.stack_config["provider"]["name"]),
                region=self.region,
                deployment_config=self.stack_config["deployment"],
            ).configure_deployment()

            KubernetesStack(
                state_file_name=self.state_file_name,
                region=self.region,
                account_id=self.account_id,
                provider=self.provider,
                deployment_type=DeploymentType.KUBERNETES,
                stacks=self.stack_config["stack"],
            ).generate()

        elif deployment_type == DeploymentType.KIND.value:
            KindDeployment(
                stack_name=self.stack_name,
                provider=CloudProvider(self.stack_config["provider"]["name"]),
                region=self.region,
                deployment_config=self.stack_config["deployment"],
            ).configure_deployment()

            LocalStack(
                state_file_name=self.state_file_name,
                region=self.region,
                account_id=self.account_id,
                provider=self.provider,
                deployment_type=DeploymentType.KIND,
                stacks=self.stack_config["stack"],
            ).generate()

        elif deployment_type == DeploymentType.MINIKUBE.value:
            MiniKubeDeployment(
                stack_name=self.stack_name,
                provider=CloudProvider(self.stack_config["provider"]["name"]),
                region=self.region,
                deployment_config=self.stack_config["deployment"],
            ).configure_deployment()

            LocalStack(
                state_file_name=self.state_file_name,
                region=self.region,
                account_id=self.account_id,
                provider=self.provider,
                deployment_type=DeploymentType.MINIKUBE,
                stacks=self.stack_config["stack"],
            ).generate()

        else:
            raise ValueError(f"Deployment type {deployment_type} not supported")

    def configure_provider(self) -> CloudProvider:
        """
        Configures the provider details based on the stack configuration.

        Returns:
            CloudProvider: The cloud provider.

        Raises:
            NotImplementedError: If the cloud provider is not supported.
        """
        provider_name = self.stack_config["provider"]["name"]
        if provider_name == CloudProvider.AWS.value:
            aws_provider = AWSProvider(
                stack_name=self.stack_name, config=self.stack_config["provider"]
            )
            aws_provider.configure_provider()
            return CloudProvider.AWS
        elif provider_name == CloudProvider.LOCAL.value:
            local_provider = LocalProvider(
                stack_name=self.stack_name,
                config=self.stack_config["provider"],
                deployment=self.stack_config["deployment"],
            )
            local_provider.configure_provider()
            return CloudProvider.LOCAL
        else:
            raise NotImplementedError("Cloud provider not supported")



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import pytest
import yaml
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.stack_processor.deployment_processor.deployment import AbstractDeployment


class TestAbstractDeployment:
    # stack_name argument is None
    def test_stack_name_argument_is_none(self):
        stack_name = None
        provider = CloudProvider.AWS
        region = "us-west-2"

        yaml_config = """
        deployment:
            type: cloud_vm
        """
        deployment_config = yaml.safe_load(yaml_config)

        with pytest.raises(TypeError):
            AbstractDeployment(stack_name, provider, region, deployment_config)

    # region argument is None
    def test_region_argument_is_none(self):
        stack_name = "test_stack"
        provider = CloudProvider.AWS
        region = None

        yaml_config = """
        deployment:
            type: cloud_vm
        """
        deployment_config = yaml.safe_load(yaml_config)

        with pytest.raises(TypeError):
            AbstractDeployment(stack_name, provider, region, deployment_config)

    # deployment_config argument is None
    def test_deployment_config_argument_is_none(self):
        stack_name = "test_stack"
        provider = CloudProvider.AWS
        region = "us-west-2"
        deployment_config = None

        with pytest.raises(TypeError):
            AbstractDeployment(stack_name, provider, region, deployment_config)



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
from importlib import resources
from typing import Any, Dict

from mlinfra import modules
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.stack_processor.deployment_processor.deployment import (
    AbstractDeployment,
)
from mlinfra.utils.utils import generate_tf_json


class CloudVMDeployment(AbstractDeployment):
    """
    A class that configures the deployment of cloud infrastructure resources based on the specified provider.

    Args:
        stack_name (str): The name of the stack.
        provider (CloudProvider): The cloud provider (AWS, GCP, or Azure).
        region (str): The region where the resources will be deployed.
        deployment_config (Dict[str, Any]): The deployment configuration in dictionary format.
    """

    def __init__(
        self,
        stack_name: str,
        provider: CloudProvider,
        region: str,
        deployment_config: Dict[str, Any],
    ):
        super().__init__(stack_name, provider, region, deployment_config)

    def configure_required_provider_config(self):
        """
        Configures the required provider configuration for the deployment.
        Updates the Terraform JSON file with the necessary provider information.
        """

        with open(
            resources.files(modules) / f"cloud/{self.provider.value}/terraform.tf.json",
            "r",
        ) as data_json:
            data = json.load(data_json)

            # add random provider
            with open(
                resources.files(modules)
                / f"cloud/{self.provider.value}/terraform_providers/random/terraform.tf.json",
                "r",
            ) as random_tf:
                random_tf_json = json.load(random_tf)
            data["terraform"]["required_providers"].update(
                random_tf_json["terraform"]["required_providers"]
            )

            data["terraform"].update(self.get_provider_backend(provider=self.provider))

            generate_tf_json(module_name="terraform", json_module=data)

    def configure_deployment_config(self):
        """
        Configures the deployment configuration based on the provider.
        Generates a Terraform JSON file for the specific provider.
        """
        if self.provider == CloudProvider.AWS:
            json_module = {
                "module": {
                    "vpc": {
                        "name": f"{self.stack_name}-vpc",
                        "source": "./modules/cloud/aws/vpc",
                    }
                }
            }
            vpc_config = self.deployment_config.get("config", {}).get("vpc", {})
            json_module["module"]["vpc"].update(vpc_config)

            generate_tf_json(module_name="vpc", json_module=json_module)
        elif self.provider == CloudProvider.GCP:
            pass
        elif self.provider == CloudProvider.AZURE:
            pass
        else:
            raise ValueError(f"Provider {self.provider} is not supported")

    def configure_deployment(self):
        """
        Configures the deployment by calling the `configure_required_provider_config()` and `configure_deployment_config()` methods.
        """
        self.configure_required_provider_config()
        self.configure_deployment_config()


if __name__ == "__main__":
    pass



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.stack_processor.deployment_processor.cloud_vm_deployment import CloudVMDeployment


class TestCloudVMDeployment:
    # can be instantiated with required parameters
    def test_instantiation(self):
        stack_name = "test_stack"
        provider = CloudProvider.AWS
        region = "us-west-2"
        deployment_config = {"config": {"vpc": {"cidr_block": "10.0.0.0/16"}}}

        deployment = CloudVMDeployment(stack_name, provider, region, deployment_config)

        assert deployment.stack_name == stack_name
        assert deployment.provider == provider
        assert deployment.region == region
        assert deployment.deployment_config == deployment_config

    # can configure deployment for AWS provider
    def test_configure_aws_deployment(self, mocker):
        stack_name = "test_stack"
        provider = CloudProvider.AWS
        region = "us-west-2"
        deployment_config = {"config": {"vpc": {"cidr_block": "10.0.0.0/16"}}}

        deployment = CloudVMDeployment(stack_name, provider, region, deployment_config)

        mocker.patch.object(deployment, "configure_required_provider_config")
        mocker.patch.object(deployment, "configure_deployment_config")

        deployment.configure_deployment()

        deployment.configure_required_provider_config.assert_called_once()
        deployment.configure_deployment_config.assert_called_once()



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import pytest
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.stack_processor.deployment_processor.kubernetes_deployment import KubernetesDeployment


class TestKubernetesDeployment:
    # can be instantiated with valid arguments
    def test_instantiation_with_valid_arguments(self):
        stack_name = "test-stack"
        provider = CloudProvider.AWS
        region = "us-west-2"
        deployment_config = {}

        deployment = KubernetesDeployment(stack_name, provider, region, deployment_config)

        assert deployment.stack_name == stack_name
        assert deployment.provider == provider
        assert deployment.region == region
        assert deployment.deployment_config == deployment_config

    def test_specified_provider_not_supported(self):
        provider = CloudProvider.GCP
        region = "us-west-2"
        stack_name = "my-stack"
        deployment_config = {
            "config": {
                "vpc": {
                    "cidr_block": "10.0.0.0/16",
                    "subnet_cidr_blocks": ["10.0.1.0/24", "10.0.2.0/24"],
                },
                "kubernetes": {
                    "cluster_version": "1.30",
                    "node_groups": [
                        {
                            "name": "worker-group",
                            "instance_type": "t3.medium",
                            "desired_capacity": 2,
                        }
                    ],
                },
            }
        }

        deployment = KubernetesDeployment(
            stack_name=stack_name,
            provider=provider,
            region=region,
            deployment_config=deployment_config,
        )

        # Assert that a FileNotFoundError is raised when the specified provider is not supported
        with pytest.raises(FileNotFoundError):
            deployment.generate_required_provider_config()



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import os
import platform
import random
import string
import sys
from typing import Optional

from getmac import get_mac_address
from git.config import GitConfigParser
from mlinfra.utils.constants import (
    DEV_VERSION,
    MLINFRA_DISABLE_REPORTING,
    SESSION_ID,
    VERSION,
)
from mlinfra.utils.utils import safe_run
from requests import codes, post


class AmplitudeClient:
    PLAN_EVENT = "PLAN"
    APPLY_EVENT = "APPLY"
    DESTROY_EVENT = "DESTROY"
    COST_ESTIMATOR_EVENT = "COST_ESTIMATOR_EVENT"
    START_GEN_TERRAFORM_EVENT = "START_GEN_TERRAFORM"
    FINISH_GEN_TERRAFORM_EVENT = "FINISH_GEN_TERRAFORM"
    VALID_EVENTS = [
        APPLY_EVENT,
        PLAN_EVENT,
        DESTROY_EVENT,
        COST_ESTIMATOR_EVENT,
        START_GEN_TERRAFORM_EVENT,
        FINISH_GEN_TERRAFORM_EVENT,
    ]

    def __init__(self) -> None:
        self.api_key = "167aefed6bd10a9200b052a52504ff5"
        self.user_id = GitConfigParser().get_value("user", "email", "no_user")
        self.device_id = get_mac_address()
        self.os_name = os.name
        self.platform = platform.system()
        self.os_version = platform.version()

    @safe_run
    def send_event(
        self,
        event_type: str,
        event_properties: Optional[dict] = None,
        user_properties: Optional[dict] = None,
    ) -> None:
        if hasattr(sys, "_called_from_test") or VERSION == DEV_VERSION:
            print("Not sending amplitude cause we think we're in a pytest or in dev")
            return
        event_properties = event_properties or {}
        user_properties = user_properties or {}
        insert_id = "".join(
            random.SystemRandom().choices(string.ascii_letters + string.digits, k=16)
        )
        if event_type not in self.VALID_EVENTS:
            raise Exception(f"Invalid event type: {event_type}")
        body = {
            "api_key": self.api_key,
            "events": [
                {
                    "user_id": self.user_id,
                    "device_id": self.device_id,
                    "event_type": event_type,
                    "event_properties": event_properties,
                    "user_properties": user_properties,
                    "app_version": VERSION,
                    "platform": self.platform,
                    "os_name": self.os_name,
                    "os_version": self.os_version,
                    "insert_id": insert_id,
                    "session_id": SESSION_ID,
                }
            ],
        }
        headers = {"Content-Type": "application/json", "Accept": "*/*"}

        if os.environ.get(MLINFRA_DISABLE_REPORTING) is None:
            try:
                r = post(
                    "https://api2.amplitude.com/2/httpapi",
                    params={},
                    headers=headers,
                    json=body,
                    timeout=10,
                )
                if r.status_code != codes.ok:
                    raise Exception(
                        "Hey, we're trying to send some analytics over to our devs for the "
                        f"product usage and we got a {r.status_code} response back. Could "
                        "you pls email over to our dev team about this and tell them of the "
                        f"failure with the aforementioned code and this response body: {r.text}"
                    )
            except Exception as err:
                print(f"Unexpected error when connecting to amplitude {err=}, {type(err)=}")


amplitude_client = AmplitudeClient()



#     Copyright (c) mlinfra 2024-2025. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from abc import ABC, abstractmethod

from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.utils.constants import TF_LOCAL_STATE_PATH


class AbstractDeployment(ABC):
    """
    Abstract class for deployment.

    Args:
        stack_name (str): The name of the deployment stack.
        provider (CloudProvider): The cloud provider for the deployment.
        region (str): The region for the deployment.
        deployment_config (dict): The deployment configuration.

    Attributes:
        stack_name (str): The name of the deployment stack.
        provider (CloudProvider): The cloud provider for the deployment.
        region (str): The region for the deployment.
        deployment_config (dict): The deployment configuration.
    """

    def __init__(
        self,
        stack_name: str,
        provider: CloudProvider,
        region: str,
        deployment_config: dict,
    ):
        self.stack_name = stack_name
        self.provider = provider
        self.region = region
        self.deployment_config = deployment_config

    @abstractmethod
    def configure_deployment(self):
        """
        Abstract method that must be implemented by subclasses to configure the deployment.
        """
        pass

    # TODO: refactor statefile name
    def get_statefile_name(self) -> str:
        """
        Get the name of the statefile for the deployment.

        Returns:
            str: The name of the statefile.
        """
        return f"tfstate-{self.stack_name}-{self.region}"

    def get_provider_backend(self, provider: CloudProvider) -> dict:
        """
        Get the backend configuration for the specified provider.

        Args:
            provider (CloudProvider): The cloud provider.

        Returns:
            json: The backend configuration.
        """
        if provider == CloudProvider.AWS:
            return {
                "backend": {
                    "s3": {
                        "bucket": self.get_statefile_name(),
                        "key": "ultimate-mlops-stack",
                        "use_lockfile": True,
                        "region": self.region,
                        "encrypt": True,
                    }
                }
            }
        elif provider == CloudProvider.LOCAL:
            return {
                "backend": {
                    "local": {
                        "path": f"{TF_LOCAL_STATE_PATH}/{self.stack_name}/terraform.tfstate",
                    }
                }
            }
        # Add other providers here
        # return {}
